import random
from collections import Counter

import ray
from dask.core import get_dependencies

from doreisa._scheduling_actor import ChunkRef


def doreisa_get(dsk, keys, **kwargs):
    # The graph may be a HighLevelGraph
    dsk = dict(dsk)

    # Sort the graph by keys to make scheduling deterministic
    dsk = {k: v for k, v in sorted(dsk.items())}

    head_node = ray.get_actor("simulation_head", namespace="doreisa")  # noqa: F841

    assert isinstance(keys, list) and len(keys) == 1
    assert isinstance(keys[0], list) and len(keys[0]) == 1
    key = keys[0][0]

    # Find the scheduling actors
    scheduling_actors = ray.get(head_node.list_scheduling_actors.remote())

    # Find a not too bad scheduling strategy
    # Good scheduling in a tree
    scheduling = {k: -1 for k in dsk.keys()}

    # def explore(key, v: int):
    #     # Only works for trees for now
    #     assert scheduling[key] == -1
    #     scheduling[key] = v
    #     for dep in get_dependencies(dsk, key):
    #         explore(dep, v)

    # scheduling[key] = 0
    # c = 0
    # for dep1 in get_dependencies(dsk, key):
    #     scheduling[dep1] = 0

    #     for dep2 in get_dependencies(dsk, dep1):
    #         scheduling[dep2] = 0

    #         for dep3 in get_dependencies(dsk, dep2):
    #             scheduling[dep3] = 0

    #             for dep4 in get_dependencies(dsk, dep3):
    #                 scheduling[dep4] = 0

    #                 for dep5 in get_dependencies(dsk, dep4):
    #                     explore(dep5, c % len(scheduling_actors))
    #                     c += 1

    # assert -1 not in scheduling.values()

    # scheduling = {k: randint(0, len(scheduling_actors) - 1) for k in dsk.keys()}
    # scheduling = {k: i % len(scheduling_actors) for i, k in enumerate(dsk.keys())}

    # Make sure the leafs are scheduled on the right actor
    # for key, val in dsk.items():
    #     match val:
    #         case ("doreisa_chunk", actor_id):
    #             scheduling[key] = actor_id
    #         case _:
    #             pass

    def explore(k) -> int:
        val = dsk[k]

        if isinstance(val, ChunkRef):
            scheduling[k] = val.actor_id
        else:
            res = [explore(dep) for dep in get_dependencies(dsk, k)]
            scheduling[k] = Counter(res).most_common(1)[0][0]

        return scheduling[k]

    explore(key)

    # Pass the scheduling to the scheduling actors
    dsk_ref, scheduling_ref = ray.put(dsk), ray.put(scheduling)  # noqa: F841

    graph_id = random.randint(0, 2**128 - 1)

    ray.get(
        [
            scheduling_actors[i]
            .schedule_graph.options(enable_task_events=False)
            .remote(dsk_ref, graph_id, scheduling_ref)
            for i in range(len(scheduling_actors))
        ]
    )

    res = scheduling_actors[scheduling[key]].get_value.remote(graph_id, key)

    return [[ray.get(ray.get(res))]]



import gc
from dataclasses import dataclass
from typing import Any, Callable

import dask.array as da
import ray

from doreisa.head_node import ArrayDefinition as HeadArrayDefinition
from doreisa.head_node import SimulationHead, get_head_actor_options


@dataclass
class ArrayDefinition:
    name: str
    window_size: int
    preprocess: Callable = lambda x: x


def run_simulation(
    simulation_callback: Callable, arrays_description: list[ArrayDefinition], *, max_iterations=1000_000_000
) -> None:
    # Convert the definitions to the type expected by the head node
    head_arrays_description = [
        HeadArrayDefinition(name=definition.name, preprocess=definition.preprocess) for definition in arrays_description
    ]
    windows_size = {definition.name: definition.window_size for definition in arrays_description}

    # Limit the advance the simulation can have over the analytics
    max_pending_arrays = 2 * len(arrays_description)

    head: Any = SimulationHead.options(**get_head_actor_options()).remote(head_arrays_description, max_pending_arrays)

    # The array values needed for the analytics
    # Each list contains the arrays for several timesteps:
    #   - The list will be shorter than the window size during the first iterations
    #   - The list may be longer than the window size if the array are not produced in order
    all_arrays: dict[str, list[da.Array]] = {description.name: [] for description in arrays_description}

    for iteration in range(max_iterations):
        # Get new arrays
        while any(len(array) < min(windows_size[name], iteration + 1) for name, array in all_arrays.items()):
            name: str
            timestep: int
            array: da.Array
            name, timestep, array = ray.get(head.get_next_array.remote())

            all_arrays[name].append(array)

        # Remove the most recent arrays if we received them in advanced
        all_arrays_cropped = {name: arrays[: windows_size[name]] for name, arrays in all_arrays.items()}

        # TODO check that the arrays were indeed produced at the same timestep
        # TODO using the last timestep might be wrong

        simulation_callback(**all_arrays_cropped, timestep=timestep)

        # Remove the oldest arrays
        for name, arrays in all_arrays.items():
            if len(arrays) >= windows_size[name]:
                arrays.pop(0)

        # Free the memory used by the arrays now. Since an ObjectRef is a small object,
        # Python may otherwise choose to keep it in memory for some time, preventing the
        # actual data to be freed.
        gc.collect()



import asyncio
import math
from dataclasses import dataclass
from typing import Callable

import dask
import dask.array as da
import numpy as np
import ray
import ray.actor
from dask.highlevelgraph import HighLevelGraph
from ray.util.scheduling_strategies import NodeAffinitySchedulingStrategy

from doreisa import Timestep
from doreisa._scheduler import doreisa_get
from doreisa._scheduling_actor import ChunkReadyInfo, ChunkRef, SchedulingActor


def init():
    if not ray.is_initialized():
        ray.init(address="auto")

    dask.config.set(scheduler=doreisa_get, shuffle="tasks")


@dataclass
class ArrayDefinition:
    """
    Description of a Dask array given by the user.
    """

    name: str
    preprocess: Callable = lambda x: x


class _DaskArrayData:
    """
    Information about a Dask array being built.
    """

    def __init__(self, definition: ArrayDefinition, timestep: Timestep) -> None:
        self.definition = definition
        self.timestep = timestep

        # This will be set when the first chunk is added
        self.nb_chunks_per_dim: tuple[int, ...] | None = None
        self.nb_chunks: int | None = None

        # For each dimension, the size of the chunks in this dimension
        self.chunks_size: list[list[int | None]] | None = None

        # Type of the numpy arrays
        self.dtype: np.dtype | None = None

        # ID of the scheduling actor in charge of the chunk at each position
        self.scheduling_actors_id: dict[tuple[int, ...], int] = {}

        # Each reference comes from one scheduling actor. The reference a list of
        # ObjectRefs, each ObjectRef corresponding to a chunk. These references
        # shouldn't be used directly. They exists only to release the memory
        # automatically.
        # When the array is buit, these references are put in the object store, and the
        # global reference is added to the Dask graph. Then, the list is cleared.
        self.chunk_refs: list[ray.ObjectRef] = []

    def add_chunk(
        self,
        size: tuple[int, ...],
        position: tuple[int, ...],
        dtype: np.dtype,
        nb_chunks_per_dim: tuple[int, ...],
        scheduling_actor_id: int,
    ) -> bool:
        """
        Add a chunk to the array.

        Return:
            True if the array is ready, False otherwise.
        """
        if self.nb_chunks_per_dim is None:
            self.nb_chunks_per_dim = nb_chunks_per_dim
            self.nb_chunks = math.prod(nb_chunks_per_dim)

            self.dtype = dtype
            self.chunks_size = [[None for _ in range(n)] for n in nb_chunks_per_dim]
        else:
            assert self.nb_chunks_per_dim == nb_chunks_per_dim
            assert self.dtype == dtype
            assert self.chunks_size is not None

        for pos, nb_chunks in zip(position, nb_chunks_per_dim):
            assert 0 <= pos < nb_chunks

        self.scheduling_actors_id[position] = scheduling_actor_id

        for d in range(len(position)):
            if self.chunks_size[d][position[d]] is None:
                self.chunks_size[d][position[d]] = size[d]
            else:
                assert self.chunks_size[d][position[d]] == size[d]

        if len(self.scheduling_actors_id) == self.nb_chunks:  # The array is ready
            return True
        return False

    def add_chunk_ref(self, chunk_ref: ray.ObjectRef) -> None:
        self.chunk_refs.append(chunk_ref)

    def get_full_array(self) -> da.Array:
        """
        Return the full Dask array.
        """
        assert len(self.scheduling_actors_id) == self.nb_chunks
        assert self.nb_chunks is not None and self.nb_chunks_per_dim is not None

        all_chunks = ray.put(self.chunk_refs)

        # We need to add the timestep since the same name can be used several times for different
        # timesteps
        dask_name = f"{self.definition.name}_{self.timestep}"

        graph = {
            # We need to repeat the name and position in the value since the key might be removed
            # by the Dask optimizer
            (dask_name,) + position: ChunkRef(
                actor_id, self.definition.name, self.timestep, position, _all_chunks=all_chunks if it == 0 else None
            )
            for it, (position, actor_id) in enumerate(self.scheduling_actors_id.items())
        }

        dsk = HighLevelGraph.from_collections(dask_name, graph, dependencies=())

        full_array = da.Array(
            dsk,
            dask_name,
            chunks=self.chunks_size,
            dtype=self.dtype,
        )

        return full_array


@dask.delayed
def ray_to_dask(x):
    return x


def get_head_actor_options() -> dict:
    """Return the options that should be used to start the head actor."""
    return dict(
        # The workers will be able to access to this actor using its name
        name="simulation_head",
        namespace="doreisa",
        # Schedule the actor on this node
        scheduling_strategy=NodeAffinitySchedulingStrategy(
            node_id=ray.get_runtime_context().get_node_id(),
            soft=False,
        ),
        # Prevents the actor from being stuck when it needs to gather many refs
        max_concurrency=1000_000_000,
        # Prevents the actor from being deleted when the function ends
        lifetime="detached",
        # Disabled for performance reasons
        enable_task_events=False,
    )


@ray.remote
class SimulationHead:
    def __init__(self, arrays_definitions: list[ArrayDefinition], max_pending_arrays: int = 1_000_000_000) -> None:
        """
        Initialize the simulation head.

        Args:
            arrays_description: Description of the arrays to be created.
            max_pending_arrays: Maximum number of arrays that can be being built or
                waiting to be collected at the same time. Setting the value can prevent
                the simulation to be many iterations in advance of the analytics.
        """

        # For each ID of a simulation node, the corresponding scheduling actor
        self.scheduling_actors: dict[str, ray.actor.ActorHandle] = {}

        self.arrays_definition: dict[str, ArrayDefinition] = {
            definition.name: definition for definition in arrays_definitions
        }

        # Must be used before creating a new array
        self.new_pending_array_semaphore = asyncio.Semaphore(max_pending_arrays)

        # Triggered when a new array is added to self.arrays
        self.new_array_created = asyncio.Event()

        # Arrays beeing built
        self.arrays: dict[tuple[str, Timestep], _DaskArrayData] = {}

        # All the newly created arrays
        self.arrays_ready: asyncio.Queue[tuple[str, int, da.Array]] = asyncio.Queue()

    def list_scheduling_actors(self) -> list[ray.actor.ActorHandle]:
        """
        Return the list of scheduling actors.
        """
        return list(self.scheduling_actors.values())

    async def scheduling_actor(self, node_id: str, *, is_fake_id: bool = False) -> ray.actor.ActorHandle:
        """
        Return the scheduling actor for the given node ID.

        Args:
            node_id: The ID of the node.
            is_fake_id: If True, the ID isn't a Ray node ID, and the actor can be scheduled
                anywhere. This is useful for testing purposes.
        """

        if node_id not in self.scheduling_actors:
            actor_id = len(self.scheduling_actors)

            if is_fake_id:
                self.scheduling_actors[node_id] = SchedulingActor.remote(actor_id)  # type: ignore
            else:
                self.scheduling_actors[node_id] = SchedulingActor.options(  # type: ignore
                    # Schedule the actor on this node
                    scheduling_strategy=NodeAffinitySchedulingStrategy(
                        node_id=node_id,
                        soft=False,
                    ),
                    max_concurrency=1000_000_000,  # Prevents the actor from being stuck
                    num_cpus=0,
                    enable_task_events=False,
                ).remote(actor_id)

            await self.scheduling_actors[node_id].ready.remote()  # type: ignore

        return self.scheduling_actors[node_id]

    def preprocessing_callbacks(self) -> dict[str, Callable]:
        """
        Return the preprocessing callbacks for each array.
        """
        return {name: definition.preprocess for name, definition in self.arrays_definition.items()}

    async def chunks_ready(
        self, chunks: list[ChunkReadyInfo], scheduling_actor_id: int, all_chunks_ref: list[ray.ObjectRef]
    ) -> None:
        """
        Called by the scheduling actors to inform the head actor that the chunks are ready.
        The chunks are not sent.

        Args:
            chunks: Information about the chunks that are ready.
            source_actor: Handle to the scheduling actor owning the chunks.
        """
        for it, chunk in enumerate(chunks):
            while (chunk.array_name, chunk.timestep) not in self.arrays:
                t1 = asyncio.create_task(self.new_pending_array_semaphore.acquire())
                t2 = asyncio.create_task(self.new_array_created.wait())

                done, pending = await asyncio.wait([t1, t2], return_when=asyncio.FIRST_COMPLETED)

                for task in pending:
                    task.cancel()

                if t1 in done:
                    if (chunk.array_name, chunk.timestep) in self.arrays:
                        # The array was already created by another scheduling actor
                        self.new_pending_array_semaphore.release()
                    else:
                        self.arrays[(chunk.array_name, chunk.timestep)] = _DaskArrayData(
                            self.arrays_definition[chunk.array_name], chunk.timestep
                        )

                        self.new_array_created.set()
                        self.new_array_created.clear()

            array = self.arrays[(chunk.array_name, chunk.timestep)]

            # TODO refactor so that the function works with only one array
            if it == 0:
                array.add_chunk_ref(all_chunks_ref[0])

            is_ready = array.add_chunk(
                chunk.size, chunk.position, chunk.dtype, chunk.nb_chunks_per_dim, scheduling_actor_id
            )

            if is_ready:
                self.arrays_ready.put_nowait(
                    (
                        chunk.array_name,
                        array.timestep,
                        array.get_full_array(),
                    )
                )
                del self.arrays[(chunk.array_name, chunk.timestep)]

    async def get_next_array(self) -> tuple[str, int, da.Array]:
        array = await self.arrays_ready.get()
        self.new_pending_array_semaphore.release()
        return array
